#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # 2A.ml - Analyse de sentiments - correction
# 
# C'est désormais un problème classique de machine learning. D'un côté, du texte, de l'autre une appréciation, le plus souvent binaire, positive ou négative mais qui pourrait être graduelle.

# In[1]:


get_ipython().run_line_magic('matplotlib', 'inline')


# In[2]:


from jyquickhelper import add_notebook_menu
add_notebook_menu()


# ## Les données
# 
# On récupère les données depuis le site UCI [Sentiment Labelled Sentences Data Set](https://archive.ics.uci.edu/ml/datasets/Sentiment+Labelled+Sentences) où on utilise la fonction ``load_sentiment_dataset``.

# In[3]:


from ensae_teaching_cs.data import load_sentiment_dataset
df = load_sentiment_dataset()
df.head()


# ## Exercice 1 : approche td-idf
# 
# La cible est la colonne *sentiment*, les deux autres colonnes sont les features. Il faudra utiliser les prétraitements [LabelEncoder](http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.LabelEncoder.html), [OneHotEncoder](http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.OneHotEncoder.html), [TF-IDF](http://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.text.TfidfVectorizer.html). L'un d'entre eux n'est pas nécessaire depuis la version [0.20.0](http://scikit-learn.org/stable/whats_new.html#sklearn-preprocessing) de *scikit-learn*. On s'occupe des variables catégorielles.

# ### La variable catégorielle
# 
# Ce serait un peu plus simple avec le module [Category Encoders](http://contrib.scikit-learn.org/categorical-encoding/) ou la dernière nouveauté de scikit-learn : [ColumnTransformer](http://scikit-learn.org/stable/modules/generated/sklearn.compose.ColumnTransformer.html#sklearn.compose.ColumnTransformer).

# In[4]:


from sklearn.model_selection import train_test_split
X_train, X_test, y_train, y_test = train_test_split(
                df.drop("sentiment", axis=1), df["sentiment"])


# In[5]:


from sklearn.preprocessing import LabelEncoder, OneHotEncoder
le = LabelEncoder()
le.fit(X_train["source"])
X_le = le.transform(X_train["source"])
X_le.shape


# In[6]:


X_le_mat = X_le.reshape((X_le.shape[0], 1))


# In[7]:


ohe = OneHotEncoder(categories="auto")
ohe.fit(X_le_mat)


# In[8]:


X_le_encoded = ohe.transform(X_le_mat)
train_cat = X_le_encoded.todense()
test_cat = ohe.transform(le.transform(X_test["source"]).reshape((len(X_test), 1))).todense()


# In[9]:


import pandas
X_train2 = pandas.concat([X_train.reset_index(drop=True),
                          pandas.DataFrame(train_cat, columns=le.classes_)],
                         sort=False, axis=1)
X_train2.head(n=2)


# In[10]:


X_test2 = pandas.concat([X_test.reset_index(drop=True),
                         pandas.DataFrame(test_cat, columns=le.classes_)],
                         sort=False, axis=1)
X_test2.head(n=2)


# ### tokenisation
# 
# On tokenise avec le module [spacy](https://spacy.io/usage/spacy-101#annotations-token) qui requiert des données supplémentaires pour découper en mot avec ``pip install https://github.com/explosion/spacy-models/releases/download/en_core_web_sm-2.0.0/en_core_web_sm-2.0.0.tar.gz`` selon les instructions dévoilées dans le [guide de départ](https://spacy.io/usage/models) ou encore ``python -m spacy download en``. Le module [gensim](http://www.xavierdupre.fr/app/papierstat/helpsphinx/notebooks/artificiel_tokenize.html?highlight=tokenisation#gensim) ne requiert pas d'installation. On peut aussi s'inspirer de l'example [word2vec pré-entraînés](http://www.xavierdupre.fr/app/papierstat/helpsphinx/notebooks/text_sentiment_wordvec.html#word2vec-pre-entraines).

# In[11]:


import spacy
nlp = spacy.load("en_core_web_sm")
# Ca marche après avoir installé le corpus correspondant
# python -m spacy download en_core_web_sm


# In[12]:


doc = nlp(X_train2.iloc[0,0])
[token.text for token in doc]


# ### tf-idf
# 
# Une fois que les mots sont tokenisé, on peut appliquer le *tf-idf*.

# In[13]:


from sklearn.feature_extraction.text import TfidfTransformer, CountVectorizer
from sklearn.pipeline import make_pipeline
tokenizer = lambda text: [token.text.lower() for token in nlp(text)]
count = CountVectorizer(tokenizer=tokenizer, analyzer='word')
tfidf = TfidfTransformer()
pipe = make_pipeline(count, tfidf)


# In[14]:


pipe.fit(X_train["sentance"])


# In[15]:


train_feature = pipe.transform(X_train2["sentance"])
train_feature


# In[16]:


test_feature = pipe.transform(X_test2["sentance"])


# ### Combinaison de toutes les variables

# In[17]:


train_feature.shape, train_cat.shape


# In[18]:


import numpy
np_train = numpy.hstack([train_feature.todense(), train_cat])
np_test = numpy.hstack([test_feature.todense(), test_cat])


# ### Calage d'un modèle

# In[19]:


from sklearn.ensemble import RandomForestClassifier
rf = RandomForestClassifier(n_estimators=50)
rf.fit(np_train, y_train)


# In[20]:


rf.score(np_test, y_test)


# In[21]:





# ## Exercice 2 : word2vec
# 
# On utilise l'approche [word2vec](https://en.wikipedia.org/wiki/Word2vec) du module [gensim](https://radimrehurek.com/gensim/models/word2vec.html) ou [spacy](https://spacy.io/usage/vectors-similarity). Avec [spacy](https://spacy.io/usage/vectors-similarity), c'est assez simple :

# In[22]:


vv = nlp(X_train2.iloc[0, 0])
list(vv)[0].vector[:10], vv.vector.shape


# On fait la somme.

# In[23]:


sum([_.vector for _ in vv])[:10]


# In[24]:


np_train_vect = numpy.zeros((X_train2.shape[0], vv.vector.shape[0]))
for i, sentance in enumerate(X_train2["sentance"]):
    np_train_vect[i, :] = sum(v.vector for v in nlp(sentance.lower()))


# In[25]:


np_test_vect = numpy.zeros((X_test2.shape[0], vv.vector.shape[0]))
for i, sentance in enumerate(X_test2["sentance"]):
    np_test_vect[i, :] = sum(v.vector for v in nlp(sentance.lower()))


# In[26]:


np_train_v = numpy.hstack([np_train_vect, train_cat])
np_test_v = numpy.hstack([np_test_vect, test_cat])


# In[27]:


rfv = RandomForestClassifier(n_estimators=50)
rfv.fit(np_train_v, y_train)


# In[28]:


rfv.score(np_test_v, y_test)


# Moins bien...

# ## Exercice 3 : comparer les deux approches
# 
# Avec une courbe [ROC](http://scikit-learn.org/stable/modules/generated/sklearn.metrics.roc_curve.html) par exemple.

# In[29]:


pmodel1 = rf.predict_proba(np_test)[:, 1]
pmodel2 = rfv.predict_proba(np_test_v)[:, 1]


# In[30]:


from sklearn.metrics import roc_auc_score, roc_curve, auc
fpr1, tpr1, th1 = roc_curve(y_test, pmodel1)
fpr2, tpr2, th2 = roc_curve(y_test, pmodel2)


# In[31]:


import matplotlib.pyplot as plt
fig, ax = plt.subplots(1,1, figsize=(4,4))
ax.plot(fpr1, tpr1, label='tf-idf')
ax.plot(fpr2, tpr2, label='word2vec')
ax.legend();


# ## Petite analyse d'erreurs
# 
# On combine les erreurs des modèles sur la base de test.

# In[32]:


final = X_test.copy()
final["model1"] = pmodel1
final["model2"] = pmodel2
final["label"] = y_test
final.head()


# On regarde des erreurs.

# In[33]:


erreurs = final[final["label"] == 1].sort_values("model2")
erreurs.head()


# In[34]:


list(erreurs["sentance"])[:5]


# Le modèle 2 reconnaît mal les négations visiblement. On regarde le modèle 1.

# In[35]:


erreurs = final[final["label"] == 1].sort_values("model1")
erreurs.head()


# In[36]:


list(erreurs["sentance"])[:5]


# Idem, voyons là où les modèles sont en désaccords.

# In[37]:


final["diff"] = final.model1 - final.model2


# In[38]:


erreurs = final[final["label"] == 1].sort_values("diff")
erreurs.head()


# In[39]:


erreurs.tail()


# Le modèle 2 (word2vec) a l'air meilleur sur les phrases longues, le modèle 1 (tf-idf) saisit mieux les mots positifs. A confirmer sur plus de données. 
# 
# * Enlever les stop words, les signes de ponctuation.
# * Combiner les deux approches.
# * n-grammes
# * ...
# 
# Dernière analyse en regardant le taux d'erreur par source.

# In[40]:


r1 = rf.predict(np_test)
r2 = rfv.predict(np_test_v)
final["rep1"] = r1
final["rep2"] = r2
final["err1"] = (final.label - final.rep1).abs()
final["err2"] = (final.label - final.rep2).abs()
final["total"] = 1
final.head()


# In[41]:


final[["source", "err1", "err2", "total"]].groupby("source").sum()


# *imdb* paraît une source une peu plus difficile à saisir. Quoiqu'il en soit, 2000 phrases pour apprendre est assez peu pour apprendre.

# ## Versions utilisées pour ce notebook
# 
# [spacy](https://spacy.io/) s'est montré quelque peu fantasques cette année avec quelques erreurs notamment celle-ci :
# [ValueError: cymem.cymem.Pool has the wrong size, try recompiling](https://github.com/explosion/spaCy/issues/2852). Voici les versions utilisées...

# In[42]:


def version(module, sub=True):
    try:
        ver = getattr(module, '__version__', None)
        if ver is None:
            ver = [_ for _ in os.listdir(os.path.join(module.__file__, '..', '..' if sub else '')) \
                   if module.__name__ in _ and 'dist' in _][-1]
        return ver
    except Exception as e:
        return str(e)


# In[43]:


import os
import thinc
print("thinc", version(thinc))
import preshed
print("preshed", version(preshed))
import cymem
print("cymem", version(cymem))
import murmurhash
print("murmurhash", version(murmurhash))
import spacy
print("spacy", spacy.__version__)

import msgpack
print("msgpack", version(msgpack))
import numpy
print("numpy", numpy.__version__)


# In[44]: